// SPDX-FileCopyrightText: 2021-2024 The Ikarus Developers mueller@ibb.uni-stuttgart.de
// SPDX-License-Identifier: LGPL-3.0-or-later

/**
 * \defgroup group_main Ikarus Modules
 * \brief All Ikarus modules
 *
 * @{
 */

/**  \defgroup assembler Assembler
\brief All finite element assembler
*/
/**  \defgroup controlroutines Control Routines
\brief Routines to follow a non-linear solution curve
 */

/**  \defgroup finiteelements Finite Elements
\brief Local assemblers for linear algebra contributions to the global system
 * @{
 */
/**  \defgroup mechanics Mechanics
\brief All finite elements in the context of mechanics
*/
/** \defgroup Affordancestags FE Affordance Tags
\brief All finite element affordances
*/
/**  \defgroup FEParameterTags FE Parameter Tags
\brief All finite element parameter tags
*/
/** @} */

/**  \defgroup materials Materials
\brief Materials for mechanical simulations
 * @{*/
/**  \defgroup Materialtags Material Tags
\brief All tags related to strain or stress types
*/
/** @} */

/**  \defgroup io IO
\brief Functions for input and output
 * @{*/
/**  \defgroup resultevaluators Result Evaluator
\brief Computation of finite element results using `calculateAt`
*/
/** @} */

/**  \defgroup pythonbindings Python bindings
\brief Header for Python Bindings
*/

/**  \defgroup solvers Solver
\brief Solvers for algebraic system of equations
*/

/**  \defgroup observer Observer
\brief Solvers for algebraic system of equations
*/

/**  \defgroup utils Utilities
\brief Collection of several utilities
*/

/**  \defgroup tensor Tensor Utilities
\brief Collection of several utilities for dealing with Eigen tensors
*/

/**  \defgroup algos Algorithms
\brief Stl-like algorithms for runtime and compile time
*/

/**  \defgroup traits Type traits
\brief Type traits for TMP
*/

/** @} */
